/* 
 Copyright (c) 2000  
 International Computer Science Institute
 All rights reserved.

 This file may contain software code originally developed for the
 Sting project. The Sting software carries the following copyright:

 Copyright (c) 1998, 1999
 Stefan Savage and the University of Washington.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgment:
      This product includes software developed by ACIRI, the AT&T
      Center for Internet Research at ICSI (the International Computer
      Science Institute). This product may also include software developed
      by Stefan Savage at the University of Washington.  
 4. The names of ACIRI, ICSI, Stefan Savage and University of Washington
    may not be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
#include "base.h"
#include "inet.h"
#include "session.h"
#include "capture.h"
#include "support.h"
#include "tbit.h"
#include "history.h"

extern struct TcpSession session;
extern struct History history[];

/*
 * Deal with struct in_addr type agreement once and for all
 */
char *InetAddress(uint32 addr)
{
  struct in_addr s;
  s.s_addr = addr;
  return (inet_ntoa(s));
}

/*
 * Really slow implementation of ip checksum
 * ripped off from rfc1071
 */
uint16 InetChecksum(uint16 *addr, uint16 len){
  uint32 sum = 0;
  uint32 count = len;

  while( count > 1 )  {
    sum += *addr++;
    count -= 2;
  }
  if( count > 0 ) {
    sum += * (uint8 *) addr;
  }

  while (sum>>16) {
    sum = (sum & 0xffff) + (sum >> 16);
  }
  return(~sum);
}


void WriteIPPacket(struct IPPacket *p,
		    uint32 src, uint32 dst, uint16 sport, uint16 dport,
		    uint32 seq, uint32 ack, uint8 flags, uint16 win,
		    uint16 urp, uint16 datalen, uint16 optlen, uint8 iptos, uint8 u4tf) 
{
  struct TcpHeader *tcp = &p->tcp;
  struct IpHeader *ip = &p->ip;

  bzero((char *)p, sizeof (struct IpHeader)); 
  ip->ip_src = src;
  ip->ip_dst = dst;
  ip->ip_p = IPPROTOCOL_TCP;
  ip->ip_xsum =
    htons((uint16)(sizeof(struct TcpHeader)+datalen+optlen)); /* pseudo hdr */
  tcp->tcp_sport = htons(sport);
  tcp->tcp_dport = htons(dport);
  tcp->tcp_seq = htonl(seq);
  tcp->tcp_ack = htonl(ack);
  tcp->tcp_hl = (sizeof (struct TcpHeader) + optlen) << 2;
  tcp->tcp_hl = tcp->tcp_hl | u4tf;
  tcp->tcp_flags = flags;
  tcp->tcp_win = htons(win);
  tcp->tcp_urp = htons(urp);
	
  tcp->tcp_xsum = 0;
  tcp->tcp_xsum = InetChecksum((uint16 *)ip, 
															 (uint16)(sizeof(struct TcpHeader) + 
																				sizeof(struct IpHeader)+ datalen+optlen));


  /* Fill in real ip header */
  ip->ip_ttl = 60;
  ip->ip_tos = iptos;
  ip->ip_vhl = 0x45;
  ip->ip_p = IPPROTOCOL_TCP;
#ifdef __FreeBSD__
  ip->ip_off = IP_DF;
  ip->ip_len = (uint16)(sizeof (struct IPPacket) + datalen + optlen);
#else /* __FreeBSD__ */
  ip->ip_off = htons(IP_DF);
  ip->ip_len = htons((uint16)((sizeof (struct IPPacket) + datalen + optlen)));
#endif /* __FreeBSD__ */

}

void ReadIPPacket(struct IPPacket *p,
		    uint32 *src, uint32 *dst, uint16 *sport, uint16 *dport,
		    uint32 *seq, uint32 *ack, uint8 *flags, uint16 *win,
		    uint16 *urp, uint16 *datalen, uint16 *optlen) 
{
  struct TcpHeader *tcp = &p->tcp;
  struct IpHeader *ip = &p->ip;

  uint16 ip_len;
  uint16 ip_hl;
  uint16 tcp_hl;

  /* XXX do checksum check? */
  if (ip->ip_p != IPPROTOCOL_TCP) {
    printf("Error: not a TCP packet\n");
		Quit(ERR_CHECKSUM);
  }

  *src = ip->ip_src;
  *dst = ip->ip_dst;
  *sport = ntohs(tcp->tcp_sport);
  *dport = ntohs(tcp->tcp_dport);
  *seq = ntohl(tcp->tcp_seq);
  *ack = ntohl(tcp->tcp_ack);
  *flags = tcp->tcp_flags;
  *win = ntohs(tcp->tcp_win);
  *urp = ntohs(tcp->tcp_urp);

  tcp_hl = tcp->tcp_hl >> 2;
  ip_len = ntohs(ip->ip_len);
  ip_hl = (ip->ip_vhl & 0x0f) << 2;
  *datalen = (ip_len - ip_hl) - tcp_hl;
  *optlen = tcp_hl - (unsigned int)sizeof(struct TcpHeader);
}

void PrintTcpPacket(struct IPPacket *p)
{
  struct TcpHeader *tcp = &p->tcp;
  struct IpHeader *ip = &p->ip;
	int i ;
	char *opt ; 
	int optlen ; 

  printf("%s.%u > ",InetAddress(ip->ip_src), ntohs(tcp->tcp_sport));
  printf("%s.%u ", InetAddress(ip->ip_dst), ntohs(tcp->tcp_dport));
  if (tcp->tcp_flags & TCPFLAGS_FIN) {
    printf("F");
  }
  if (tcp->tcp_flags & TCPFLAGS_SYN) {
    printf("S");
  }
  if (tcp->tcp_flags & TCPFLAGS_RST) {
    printf("R");
  }
  if (tcp->tcp_flags & TCPFLAGS_PSH) {
    printf("P");
  }
  if (tcp->tcp_flags & TCPFLAGS_ACK) {
    printf("A");
  }
  if (tcp->tcp_flags & TCPFLAGS_URG) {
    printf("U");
  }
	if (INSESSION(p,session.src,session.sport,session.dst,session.dport)) {
  	printf(" seq: %u, ack: %u", (uint32)ntohl(tcp->tcp_seq)-session.iss, (uint32)ntohl(tcp->tcp_ack)-session.irs);
	}
	else {
  	printf(" seq: %u, ack: %u", (uint32)ntohl(tcp->tcp_seq)-session.irs, (uint32)ntohl(tcp->tcp_ack)-session.iss);
	}
  printf(" win: %u, urg: %u",ntohs(tcp->tcp_win), ntohs(tcp->tcp_urp));
  printf(" datalen: %u, optlen: %u ",
	 ntohs(ip->ip_len) - ((ip->ip_vhl &0x0f) << 2) - (tcp->tcp_hl >> 2),
	 (tcp->tcp_hl >> 2) - (unsigned int)sizeof (struct TcpHeader));
 
  optlen = (tcp->tcp_hl >> 2) - (unsigned int)sizeof (struct TcpHeader) ;
	opt = (char *)p+sizeof(struct IPPacket) ;
	i = 0 ;
	while (i < optlen) {
		switch ((unsigned char)opt[i]) {
			case TCPOPT_EOL: 
				printf (" opt%d: %s ", i+1, "TCPOPT_EOL");
				i = optlen + 1; 
				break ; 
			case TCPOPT_NOP:
				printf (" opt%d: %s ", i+1, "TCPOPT_NOP");
				i++ ; 
				break ;
			case TCPOPT_MAXSEG:
				printf (" opt%d: %s: %d ", i+1, "TCPOPT_MAXSEG", ntohs(*(uint16 *)((char *)opt+2)));
				i = i + TCPOLEN_MAXSEG ; 
				break ;
			case TCPOPT_WINDOW:
				printf (" opt%d: %s ", i+1, "TCPOPT_WINDOW");
				i = i + TCPOLEN_WINDOW ;
				break ; 
			case TCPOPT_SACK_PERMITTED:
				printf (" opt%d: %s ", i+1, "TCPOPT_SACK_PERMITTED");
				i = i + TCPOLEN_SACK_PERMITTED ; 
				break ; 
			case TCPOPT_TIMESTAMP:
				printf (" opt%d: %s ", i+1, "TCPOPT_TIMESTAMP");
				i = i + TCPOLEN_TIMESTAMP ; 
				break ; 
			default: 
				printf (" opt%d l%d: unknown opt ", (uint8)opt[i], (uint8)opt[i+1]);
				if ((uint8)opt[i+1] > 0) {
					i = i + (uint8)opt[i+1] ;
				} else {
				 Quit(20); 
				}
				break ;
		} 
	}
  printf ("\n");
}




