/* 
 Copyright (c) 2000  
 International Computer Science Institute
 All rights reserved.

 This file may contain software code originally developed for the
 Sting project. The Sting software carries the following copyright:

 Copyright (c) 1998, 1999
 Stefan Savage and the University of Washington.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgment:
      This product includes software developed by ACIRI, the AT&T
      Center for Internet Research at ICSI (the International Computer
      Science Institute). This product may also include software developed
      by Stefan Savage at the University of Washington.  
 4. The names of ACIRI, ICSI, Stefan Savage and University of Washington
    may not be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
#include "base.h"
#include "inet.h"
#include "session.h"
#include "capture.h"
#include "support.h"
#include "tbit.h"

struct TcpSession session;

int EstablishSession(uint32 sourceAddress, uint16 sourcePort, uint32 targetAddress, 
                     uint16 targetPort, int mss, int optlen, char *opt, int maxwin, 
										 int maxpkts, uint8 iptos)
{
	int rawSocket;
	struct IPPacket *p;
	struct IPPacket *synPacket;
	struct PacketInfo pi;
	int synAckReceived = 0;
	int numRetransmits = 0;
	double timeoutTime;
	double ts1, ts2;
	int flag=1;
	
	session.src = sourceAddress;
	session.sport = sourcePort;
	session.dst = targetAddress;
	session.dport = targetPort;
	session.rcv_wnd = maxwin*mss;
	session.snd_nxt = (uint32)mrand48();	/* random initial sequence number */
	session.iss = session.snd_nxt;
	session.rcv_nxt = 0;
	session.irs = 0;
	session.mss = mss ;
	session.maxseqseen = 0 ; 
	session.epochTime = GetTime ();
	session.maxpkts = maxpkts; 

	if ((session.dataRcvd = (uint8 *)calloc(sizeof(uint8), mss*session.maxpkts)) == NULL) {
		perror("no memmory to store data:\n");
		Quit(ERR_MEM_ALLOC);
	}

	/* Now open a raw socket for sending our "fake" TCP segments */
	if ((rawSocket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("ERROR: couldn't open socket:");
		Quit(ERR_SOCKET_OPEN);
	}
	if (setsockopt(rawSocket, IPPROTO_IP, IP_HDRINCL, (char *)&flag,sizeof(flag)) < 0) {
		perror("ERROR: couldn't set raw socket options:");
		Quit(ERR_SOCKOPT);
	}

	session.socket = rawSocket;

	/* set firewall */
	SetFireWall ();
	
	/* allocate the syn packet */
	if ((synPacket = (struct IPPacket *)calloc(1, sizeof(struct IPPacket)+optlen)) == NULL) {
		perror("ERROR: Could not allocate SYN packet:") ;
		Quit(ERR_MEM_ALLOC) ; 
	}

	/* fill in the options */
	if (optlen > 0) {
		memcpy((char *)synPacket+sizeof(struct IPPacket), opt, optlen);
	}

	/* send it */
	SendSessionPacket(synPacket, sizeof(struct IPPacket)+optlen, TCPFLAGS_SYN, optlen, iptos);	 

	timeoutTime = GetTime() + SYNTIMEOUT;

	/* 
	 * Wait for SYN/ACK and retransmit SYN if appropriate 
	 * not great, but it gets the job done 
	 */

	while(!synAckReceived && numRetransmits < MAXSYNRETRANSMITS) {
		while(GetTime() < timeoutTime) {
			/* Have we captured any packets? */
			if ((p = (struct IPPacket *)CaptureGetPacket(&pi)) != NULL) {
				/* Received a packet from us to them */
				if (INSESSION(p, session.src, session.sport, session.dst, session.dport)) {
					/* Is it a SYN? */
					if (p->tcp.tcp_flags == TCPFLAGS_SYN) {
						if (session.debug) {
							PrintTcpPacket(p); 
						}
						StorePacket(p);
						ts1 = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
						session.totSeenSent ++ ;
					}
					else {
						processBadPacket(p);
					}
					continue;
				}
				if (INSESSION(p, session.dst, session.dport, session.src, session.sport)) {
					/* Is it a SYN/ACK? */
					if (p->tcp.tcp_flags == (TCPFLAGS_SYN | TCPFLAGS_ACK)) {
						timeoutTime = GetTime(); /* force exit */
						synAckReceived++;
						ts2 = pi.ts.tv_sec + (double)pi.ts.tv_usec/1000000.0;
						session.rtt = ts2-ts1 ;
						if (numRetransmits > 0) {			
							session.rtt_unreliable = 1;
							printf("##### Unreliable\n");	/* ACK for which SYN? */
						}
						if (session.debug) {
							PrintTcpPacket(p);
							printf("Connection setup took %d ms\n",(int)((ts2-ts1)*1000.0));
						}
						StorePacket(p);
						/* Save ttl for,admittedly poor,indications of reverse route change */
						session.ttl = p->ip.ip_ttl;
						session.snd_wnd = ntohl(p->tcp.tcp_win);
						session.totRcvd ++;
						break ;
					}
					else {
						processBadPacket(p); 
					}
				}
				else {
					processBadPacket(p);
				} 
			}
		}
		if (!synAckReceived) {
			if (session.debug) {
				printf("SYN timeout.	Retransmitting\n");
			}
			SendSessionPacket(synPacket, sizeof(struct IPPacket)+optlen, TCPFLAGS_SYN, optlen, iptos);
			timeoutTime = GetTime() + SYNTIMEOUT;
			numRetransmits++;
		}
	}
	if (numRetransmits >= MAXSYNRETRANSMITS) {
		printf("ERROR: Could not establish contact after %d retries\n", numRetransmits);
		Quit(NO_CONNECTION);
	}
	
	/* Update session variables */
	session.irs = ntohl(p->tcp.tcp_seq);
	session.dataRcvd[0] = 1 ;
	session.rcv_nxt = session.irs+1;	/* SYN/ACK takes up a byte of seq space */
	session.snd_nxt = session.iss+1;	/* SYN takes up a byte of seq space */
	session.snd_una = session.iss+1;
	session.maxseqseen = ntohl(p->tcp.tcp_seq);

	session.initSession = 1;
	if (session.debug) {
		printf("src = %s:%d (%u)\n", InetAddress(session.src), session.sport, session.iss);
		printf("dst = %s:%d (%u)\n",InetAddress(session.dst), session.dport, session.irs);
	}
	free (synPacket);
	return 1;
}

int PrepareRequest(char *data, char *filename) 
{
	char h1[]= "GET /" ;
	char h2[]=" HTTP/1.0" ;
	char h3[]="Host: ";
	char h4[]="Accept: */*\nUser-Agent: Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; TBIT)";
	char deffile[]=DEFAULT_FILENAME;

	if (filename == NULL) {
		filename = deffile;
	}
	if (strlen(session.targetName) > 0) {
		sprintf(data, "%s%s%s\n%s%s\n%s\n\n", h1, filename, h2, h3, session.targetName, h4);
	}
	else {
		sprintf(data, "%s%s%s\n%s\n\n", h1, filename, h2, h4);
	}
	return strlen(data); 
}

void SendRequest(char *filename, void (*ackData)(struct IPPacket *p)) 
{
	struct IPPacket *p, *datapkt;
	struct PacketInfo pi;
	int i ;
	int sendflag = 1 ;
	double startTime;
	char *dataptr ; 
	char data[MAXREQUESTLEN];
	int datalen;
	int ipsz; 

	datalen = PrepareRequest (data, filename);
	ipsz = sizeof(struct IPPacket) + datalen + 1; 
	
	if (((char *)datapkt=(char *)calloc(ipsz, 1)) == NULL) {
		perror("ERROR: no space for request packet:");
		Quit(ERR_MEM_ALLOC) ; 
	}
	dataptr = (char *)datapkt + sizeof(struct IPPacket);
	memcpy((void *)dataptr,(void *)data, datalen);
	
	/* send the data packet
	 * we try to "achieve" reliability by
	 * sending the packet upto 5 times, wating for
	 * 2 seconds between packets
	 * BAD busy-wait loop
	 */

	i = 0 ;
	while(1) {
		if (sendflag == 1) {
			SendSessionPacket(datapkt, ipsz, TCPFLAGS_PSH | TCPFLAGS_ACK, 0, 0);
			startTime = GetTime();	
			sendflag = 0 ; 
			i ++ ;
		}
		/* Check if we have received any packets */
		if ((p =(struct IPPacket *)CaptureGetPacket(&pi)) != NULL) {
			/*
			 * packet that we sent?
			 */
			if (INSESSION(p,session.src,session.sport,session.dst,session.dport) &&
				 (p->tcp.tcp_flags == (TCPFLAGS_PSH | TCPFLAGS_ACK)) &&
				 (ntohl(p->tcp.tcp_seq) == session.snd_nxt) &&
				 (ntohl(p->tcp.tcp_ack) <= session.rcv_nxt)) {
				if (session.debug) {
					printf("saw the packet we sent: xmit %d\n", i);
					PrintTcpPacket(p);
				}
				StorePacket(p);
				session.snd_nxt += datalen + 1;
				session.totSeenSent ++ ;
				continue ;
			} 
			/*
			 * from them? 
			 */ 
			if (INSESSION(p,session.dst,session.dport,session.src,session.sport) &&
					(p->tcp.tcp_flags & TCPFLAGS_ACK) &&
					(ntohl(p->tcp.tcp_seq) == session.rcv_nxt) &&
					(ntohl(p->tcp.tcp_ack) >= session.snd_una)) {
				session.snd_una = ntohl(p->tcp.tcp_ack) ;
				if (p->ip.ip_ttl != session.ttl) {
					printf("#### WARNING: route may have changed (ttl was %d, is	%d).\n",
									session.ttl, p->ip.ip_ttl);
					session.ttl = p->ip.ip_ttl;
				}
				if (session.debug) {
					PrintTcpPacket(p);
				}
				StorePacket(p);
				session.totRcvd ++;
				/* if the packet also contains data, recieve it, and send an ack if needed */
				(*ackData)(p);
				break ;
			}
			/* 
			 * otherwise, this is a bad packet
			 * we must quit
			 */
			 processBadPacket(p);
		}
		if ((GetTime()-startTime>=REXMITDELAY) &&
				(sendflag == 0) && (i < MAXDATARETRANSMITS)) {
			sendflag = 1 ;
		}
		if (i >= MAXDATARETRANSMITS) {
			printf ("ERROR: sent request 5 times without response\n");
			Quit(NO_CONNECTION);
		}
	}	
}

void SendSessionPacket(struct IPPacket *p,
		       uint16 ip_len, uint8 tcp_flags, uint16 optlen, uint8 iptos)
{
  (void) WriteIPPacket(p,
			session.src,
			session.dst,
			session.sport,
			session.dport,
			session.snd_nxt, 
			session.rcv_nxt, 
			tcp_flags,
			session.rcv_wnd,
			0, ip_len - sizeof(struct IPPacket) - optlen, optlen, iptos, 0);
	SendPkt(p, ip_len);
}

void SendPkt(struct IPPacket *p, uint16 ip_len) {
  int nbytes;
  struct sockaddr_in sockAddr;   

  sockAddr.sin_family  = AF_INET;
  sockAddr.sin_addr.s_addr = session.dst;
  if ((nbytes = sendto(session.socket, (char *)p, ip_len, 0,
		       (struct sockaddr *)&sockAddr,
		       sizeof(sockAddr))) < ip_len) {
    printf("#### WARNING: only sent %d of %d bytes\n", nbytes, ip_len);
		perror("here");
  }
	session.totSent ++ ;
}

void rcvData (void (*ackData)(struct IPPacket *p))
{
	struct PacketInfo pi;
	struct IPPacket *p ;
	double startTime = GetTime () ;

	while (1) {
		if ((GetTime()-startTime) > (MAXDATARETRANSMITS*REXMITDELAY)) {
			printf ("####no data received for %f seconds\n", (MAXDATARETRANSMITS*REXMITDELAY));
			Quit(NO_DATA_RCVD) ;
		}
		if ((p = (struct IPPacket *)CaptureGetPacket(&pi)) != NULL) {
			/*
			 * packet that we sent?
			 */
			if (INSESSION(p,session.src,session.sport,session.dst,session.dport) &&
				 ((p->tcp.tcp_flags & TCPFLAGS_ACK) || (p->tcp.tcp_flags & TCPFLAGS_FIN)) &&
				 (ntohl(p->tcp.tcp_seq) == session.snd_nxt) &&
				 (ntohl(p->tcp.tcp_ack) <= session.rcv_nxt)) {
				if (session.debug) {
					printf("saw the packet we sent\n");
					PrintTcpPacket(p);
				}
				StorePacket(p);
				session.totSeenSent ++ ;
				continue ;
			} 
			/*
			 * data that we were expecting?
			 */ 
			if (INSESSION(p,session.dst,session.dport,session.src,session.sport) &&
					(p->tcp.tcp_flags & TCPFLAGS_ACK) &&
					//(ntohl(p->tcp.tcp_seq) >= session.rcv_nxt) &&
					(ntohl(p->tcp.tcp_ack) >= session.snd_una)) {
				if (p->ip.ip_ttl != session.ttl) {
					printf("#### WARNING: route may have changed (ttl was %d, is	%d).\n",
									session.ttl, p->ip.ip_ttl);
					session.ttl = p->ip.ip_ttl;
				}
				if (session.debug) {
					PrintTcpPacket(p);
				}
				session.totRcvd ++ ;
				startTime = GetTime () ;
				StorePacket(p);
				/* if the packet also contains data, recieve it, and send an ack if needed */
				ackData (p);
				continue ;
			}
			/* 
			 * otherwise, this is a bad packet
			 * we must quit
			 */
			 else 
				processBadPacket(p);
		}
	}
}

