/* 
 Copyright (c) 2000  
 International Computer Science Institute
 All rights reserved.

 This file may contain software code originally developed for the
 Sting project. The Sting software carries the following copyright:

 Copyright (c) 1998, 1999
 Stefan Savage and the University of Washington.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgment:
      This product includes software developed by ACIRI, the AT&T
      Center for Internet Research at ICSI (the International Computer
      Science Institute). This product may also include software developed
      by Stefan Savage at the University of Washington.  
 4. The names of ACIRI, ICSI, Stefan Savage and University of Washington
    may not be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY ICSI AND CONTRIBUTORS ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL ICSI OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.
*/
#include "base.h"
#include "inet.h"
#include "session.h"
#include "capture.h"
#include "support.h"
#include "tbit.h"
#include "reno.h"

extern struct TcpSession session; 

#ifdef __FreeBSD__
struct ip_fw firewallRule;
#endif /* __FreeBSD__ */
#ifdef linux
struct ip_fwchange firewallRule;
#endif /* linux */

void SendReset()
{
	struct IPPacket p;
	int i;

	for (i=0;i< MAXRESETRETRANSMITS;i++) {
		SendSessionPacket(&p, sizeof(struct IPPacket), TCPFLAGS_RST,0,0);
	}
}

/* make a clean exit on interrupts */
void  SigHandle (int signo)
{
	Cleanup () ; 
	fflush(stdout); 
	fflush(stderr); 
	exit(-1);
}

void Cleanup()
{
	/* If a firewall rule has been installed then remove it */
	if (session.initFirewall > 0) {
#ifdef linux
#define IP_FW_DEL	(IP_FW_DELETE)
#endif /* linux */
		if (setsockopt(session.socket,IPPROTO_IP, IP_FW_DEL, &firewallRule, sizeof(firewallRule)) != 0) {
			printf("ERROR: couldn't remove firewall rule\n");
		}
	}
	if (session.initSession > 0) {
		SendReset();
		shutdown(session.socket,2);
	}
	if (session.initCapture > 0) {
		CaptureEnd();
	}
}

void Quit(int how)
{
	Cleanup();
	fflush(stdout);
	fflush(stderr);
	exit(how);
}


double GetTime()
{
	struct timeval tv;
	struct timezone tz;
	double postEpochSecs;

	if (gettimeofday(&tv, &tz) < 0) {
		perror("GetTime");
		exit(-1);
	}
	
	postEpochSecs = (double)tv.tv_sec + ((double)tv.tv_usec/(double)1000000.0);
	return postEpochSecs;
}

void PrintTimeStamp(struct timeval *ts)
{
	(void)printf("%02d:%02d:%02d.%06u ",
				 (unsigned int)ts->tv_sec / 3600,
				 ((unsigned int)ts->tv_sec % 3600) / 60,
				 (unsigned int)ts->tv_sec % 60, (unsigned int)ts->tv_usec);
}

void processBadPacket (struct IPPacket *p)
{
	/*
	 * reset? the other guy does not like us?
	 */
	if (INSESSION(p,session.dst,session.dport,session.src,session.sport) && (p->tcp.tcp_flags & TCPFLAGS_RST)) {
		printf("ERROR: Early Reset \n");
		Quit(EARLY_RST);
	}
	/*
	 * some other packet between us that is none of the above
	 */
	if (INSESSION(p,session.src,session.sport,session.dst,session.dport) ||
		  INSESSION(p,session.dst,session.dport,session.src,session.sport)) {
		printf("ERROR: Unexpected packet \n");
		printf("Expecting:\n");
		printf("\tsrc = %s:%d (seq=%u, ack=%u)\n",
					 InetAddress(session.src), session.sport,
					 session.snd_nxt-session.iss,
					 session.rcv_nxt-session.irs);
		printf("\tdst = %s:%d (seq=%u, ack=%u)\n",
					 InetAddress(session.dst),session.dport,
					 session.rcv_nxt-session.irs, session.snd_una-session.iss);
		printf("Received:\n\t");
		PrintTcpPacket(p);
		printf ("session.snd_nxt=%d, session.rcv_nxt=%d, session.snd_una=%d\n", 
						 session.snd_nxt-session.iss, session.rcv_nxt-session.irs, session.snd_una-session.iss);
		Quit(UNEXPECTED_PKT);
	}
	/*
	 * none of the above, 
	 * so we must be seeing packets 
	 * from some other flow!
	 */
	else {
		printf("ERRROR: Received packet from different flow\n");
		PrintTcpPacket(p);
		Quit(DIFF_FLOW) ;
	}
}
/* 
 * Block kernel TCP from seeing any of the target's responses.
 * AFAIK its irrelvant that this is the same raw socket that we're
 * sending on.
 *
*/
void SetFireWall () 
{
	memset(&firewallRule, 0, sizeof(firewallRule));
#ifdef __FreeBSD__
	firewallRule.fw_flg |= IP_FW_F_DENY | IP_FW_F_IN; 
	firewallRule.fw_prot = IPPROTO_TCP;
	firewallRule.fw_src.s_addr = session.dst;
	firewallRule.fw_smsk.s_addr = htonl(~0);
	firewallRule.fw_uar.fw_pts[0] = session.dport;
	IP_FW_SETNSRCP(&firewallRule, 1);
	firewallRule.fw_dst.s_addr = session.src;
	firewallRule.fw_uar.fw_pts[1] = session.sport;
	IP_FW_SETNDSTP(&firewallRule, 1);
	firewallRule.fw_dmsk.s_addr = htonl(~0);
	firewallRule.fw_number = 999;
	if (setsockopt(session.socket, IPPROTO_IP, IP_FW_ADD, &firewallRule, sizeof(firewallRule)) != 0) {
		printf("ERROR: couldn't block kernel TCP for %s:%d\n", InetAddress(session.dst), session.dport);
	}
#endif /* __FreeBSD__ */
#ifdef linux
	memcpy(firewallRule.fwc_label, IP_FW_LABEL_INPUT, sizeof(firewallRule.fwc_label));
	memcpy(firewallRule.fwc_rule.label, IP_FW_LABEL_BLOCK, sizeof(firewallRule.fwc_label)); 
	firewallRule.fwc_rule.ipfw.fw_proto = IPPROTO_TCP;

	firewallRule.fwc_rule.ipfw.fw_src.s_addr = session.dst;
	firewallRule.fwc_rule.ipfw.fw_smsk.s_addr = htonl(~0); 
	firewallRule.fwc_rule.ipfw.fw_spts[0] = session.dport;
	firewallRule.fwc_rule.ipfw.fw_spts[1] = session.dport;
	firewallRule.fwc_rule.ipfw.fw_dst.s_addr = session.src;
	firewallRule.fwc_rule.ipfw.fw_dmsk.s_addr = htonl(~0); 
	firewallRule.fwc_rule.ipfw.fw_dpts[0] = session.sport;
	firewallRule.fwc_rule.ipfw.fw_dpts[1] = session.sport;
	if (setsockopt(session.socket,IPPROTO_IP,IP_FW_APPEND, &firewallRule,sizeof(firewallRule)) != 0) {
		printf("ERROR: couldn't block kernel TCP for %s:%d\n", InetAddress(session.dst),	session.dport);
		Quit(FIREWALL_ERR);
	}
#endif /* linux */
	session.initFirewall=1;
}

void busy_wait (double wait)
{
	double now = GetTime();
	double x = now ;
	while ((x - now) < wait) {
		x=GetTime();
	}
}
